#include "DeferredShader.h"
#include "FrameBufferObject.h"
#include "RenderState.h"
#include "ShadowMapping.h"


using namespace std;


namespace CHCDemoEngine
{


static void PrintGLerror(char *msg)
{
	GLenum errCode;
	const GLubyte *errStr;
	
	if ((errCode = glGetError()) != GL_NO_ERROR) 
	{
		errStr = gluErrorString(errCode);
		fprintf(stderr,"OpenGL ERROR: %s: %s\n", errStr, msg);
	}
}


static GLenum mymrt[] = {GL_COLOR_ATTACHMENT0_EXT, GL_COLOR_ATTACHMENT1_EXT, GL_COLOR_ATTACHMENT2_EXT,  GL_COLOR_ATTACHMENT3_EXT};


static CGprogram sCgDeferredProgram;
static CGprogram sCgDeferredShadowProgram;
static CGprogram sCgAntiAliasingProgram;


static CGparameter sColorsTexParam;
static CGparameter sPositionsTexParam;
static CGparameter sNormalsTexParam;

static CGparameter sColorsTexAntiAliasingParam;
static CGparameter sNormalsTexAntiAliasingParam;

static CGparameter sShadowMapParam;
static CGparameter sPositionsTexShadowParam;  
static CGparameter sColorsTexShadowParam;  
static CGparameter sNormalsTexShadowParam;

static CGparameter sShadowMatrixParam;




DeferredShader::DeferredShader(int w, int h):
mWidth(w), mHeight(h)
{
	//mFbo = new FrameBufferObject(w, h, FrameBufferObject::DEPTH_NONE);
	//mFbo->AddColorBuffer(ColorBufferObject::BUFFER_FLOAT_32, ColorBufferObject::WRAP_CLAMP_TO_EDGE, ColorBufferObject::FILTER_LINEAR, false);
}


DeferredShader::~DeferredShader()
{
	if (sCgDeferredProgram)
		cgDestroyProgram(sCgDeferredProgram);

	//DEL_PTR(mFbo);
}


void DeferredShader::Init(CGcontext context)
{
	sCgDeferredProgram = 
		cgCreateProgramFromFile(context, 
								CG_SOURCE,
								"src/shaders/deferred.cg", 
								RenderState::sCgFragmentProfile,
								"main",
								NULL);

	if (sCgDeferredProgram != NULL)
	{
		cgGLLoadProgram(sCgDeferredProgram);

		// we need size of texture for scaling
		sPositionsTexParam = cgGetNamedParameter(sCgDeferredProgram, "positions");  
		sColorsTexParam = cgGetNamedParameter(sCgDeferredProgram, "colors");  
		sNormalsTexParam = cgGetNamedParameter(sCgDeferredProgram, "normals"); 
	}
	else
		cerr << "deferred program failed to load" << endl;

	sCgAntiAliasingProgram = 
		cgCreateProgramFromFile(context, 
								CG_SOURCE,
								"src/shaders/antialiasing.cg", 
								RenderState::sCgFragmentProfile,
								"main",
								NULL);

	if (sCgAntiAliasingProgram != NULL)
	{
		cgGLLoadProgram(sCgAntiAliasingProgram);

		sColorsTexAntiAliasingParam = cgGetNamedParameter(sCgAntiAliasingProgram, "colors");  
		sNormalsTexAntiAliasingParam = cgGetNamedParameter(sCgAntiAliasingProgram, "normals");
	}
	else
		cerr << "antialiasing program failed to load" << endl;

	sCgDeferredShadowProgram = 
		cgCreateProgramFromFile(context, 
								CG_SOURCE,
								"src/shaders/deferred.cg", 
								RenderState::sCgFragmentProfile,
								"main_shadow",
								NULL);

	if (sCgDeferredShadowProgram != NULL)
	{
		cgGLLoadProgram(sCgDeferredShadowProgram);

		// we need size of texture for scaling
		sPositionsTexShadowParam = cgGetNamedParameter(sCgDeferredShadowProgram, "positions");  
		sColorsTexShadowParam = cgGetNamedParameter(sCgDeferredShadowProgram, "colors");  
		sNormalsTexShadowParam = cgGetNamedParameter(sCgDeferredShadowProgram, "normals"); 
		sShadowMapParam = cgGetNamedParameter(sCgDeferredShadowProgram, "shadowMap");  

		sShadowMatrixParam = cgGetNamedParameter(sCgDeferredShadowProgram, "shadowMatrix");
	}
	else
		cerr << "deferred program failed to load" << endl;

	PrintGLerror("init");
}


void DeferredShader::Render(FrameBufferObject *fbo)
{
	cgGLEnableProfile(RenderState::sCgFragmentProfile);

	glDisable(GL_ALPHA_TEST);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_LIGHTING);

	glPushAttrib(GL_VIEWPORT_BIT);
	glViewport(0, 0, mWidth, mHeight);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	const float offs = 0.5f;
	glOrtho(-offs, offs, -offs, offs, 0, 1);


	FirstPass(fbo);
	AntiAliasing(fbo);

	glEnable(GL_LIGHTING);
	glDisable(GL_TEXTURE_2D);
	
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	glPopAttrib();

	cgGLDisableProfile(RenderState::sCgFragmentProfile);
}


void DeferredShader::Render(FrameBufferObject *fbo, 
							const Matrix4x4 &matViewing, 
							ShadowMapping *shadowMapping)
{
	cgGLEnableProfile(RenderState::sCgFragmentProfile);

	glDisable(GL_ALPHA_TEST);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_LIGHTING);

	glPushAttrib(GL_VIEWPORT_BIT);
	glViewport(0, 0, mWidth, mHeight);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	const float offs = 0.5f;
	glOrtho(-offs, offs, -offs, offs, 0, 1);

	FirstPassShadow(fbo, matViewing, shadowMapping);
	AntiAliasing(fbo);

	glEnable(GL_LIGHTING);
	glDisable(GL_TEXTURE_2D);
	
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	glPopAttrib();

	cgGLDisableProfile(RenderState::sCgFragmentProfile);
}


void DeferredShader::FirstPass(FrameBufferObject *fbo)
{
	GLuint colorsTex = fbo->GetColorBuffer(0)->GetTexture();
	GLuint positionsTex = fbo->GetColorBuffer(1)->GetTexture();
	GLuint normalsTex = fbo->GetColorBuffer(2)->GetTexture();

	fbo->Bind();
	glDrawBuffers(1, mymrt + 3);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	cgGLEnableProfile(RenderState::sCgFragmentProfile);
	cgGLBindProgram(sCgDeferredProgram);

	cgGLSetTextureParameter(sColorsTexParam, colorsTex);
	cgGLEnableTextureParameter(sColorsTexParam);

	cgGLSetTextureParameter(sPositionsTexParam, positionsTex);
	cgGLEnableTextureParameter(sPositionsTexParam);

	cgGLSetTextureParameter(sNormalsTexParam, normalsTex);
	cgGLEnableTextureParameter(sNormalsTexParam);
	
	glColor3f(1.0f, 1.0f, 1.0f);
	
	glBegin(GL_QUADS);

	float offs2 = 0.5f;

	glTexCoord2f(0, 0); glVertex3f(-offs2, -offs2, -0.5f);
	glTexCoord2f(1, 0); glVertex3f( offs2, -offs2, -0.5f);
	glTexCoord2f(1, 1); glVertex3f( offs2,  offs2, -0.5f);
	glTexCoord2f(0, 1); glVertex3f(-offs2,  offs2, -0.5f);

	glEnd();

	cgGLDisableTextureParameter(sColorsTexParam);
	cgGLDisableTextureParameter(sPositionsTexParam);
	cgGLDisableTextureParameter(sNormalsTexParam);

	cgGLDisableProfile(RenderState::sCgFragmentProfile);
	
	FrameBufferObject::Release();

	PrintGLerror("deferred shading");
}


static void SetVertex(float x, float y, float x_offs, float y_offs)
{
	glMultiTexCoord2fARB(GL_TEXTURE0_ARB, x, y); // center
	glMultiTexCoord2fARB(GL_TEXTURE1_ARB, x - x_offs, y + y_offs); // left top
	glMultiTexCoord2fARB(GL_TEXTURE2_ARB, x + x_offs, y - y_offs); // right bottom
	glMultiTexCoord2fARB(GL_TEXTURE3_ARB, x + x_offs, y + y_offs); // right top
	glMultiTexCoord2fARB(GL_TEXTURE4_ARB, x - x_offs, y - y_offs); // left bottom

	glMultiTexCoord4fARB(GL_TEXTURE5_ARB, x - x_offs, y, x + x_offs, y); // left right
	glMultiTexCoord4fARB(GL_TEXTURE6_ARB, x, y + y_offs, x, y - y_offs); // top bottom

	glVertex3f(x - 0.5f, y - 0.5f, -0.5f);
}


void DeferredShader::AntiAliasing(FrameBufferObject *fbo)
{
	GLuint colorsTex = fbo->GetColorBuffer(3)->GetTexture();
	GLuint normalsTex = fbo->GetColorBuffer(2)->GetTexture();

	
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	cgGLEnableProfile(RenderState::sCgFragmentProfile);

	cgGLBindProgram(sCgAntiAliasingProgram);

	cgGLSetTextureParameter(sColorsTexAntiAliasingParam, colorsTex);
	cgGLEnableTextureParameter(sColorsTexAntiAliasingParam);

	cgGLSetTextureParameter(sNormalsTexAntiAliasingParam, normalsTex);
	cgGLEnableTextureParameter(sNormalsTexAntiAliasingParam);
	
	glColor3f(1.0f, 1.0f, 1.0f);

	float offs2 = 0.5f;

	glBegin(GL_QUADS);

	// the neighbouring texels
	float x_offs = 1.0f / mWidth;
	float y_offs = 1.0f / mHeight;

	SetVertex(0, 0, x_offs, y_offs);
	SetVertex(1, 0, x_offs, y_offs);
	SetVertex(1, 1, x_offs, y_offs);
	SetVertex(0, 1, x_offs, y_offs);

	glEnd();

	cgGLDisableTextureParameter(sColorsTexAntiAliasingParam);
	cgGLDisableTextureParameter(sNormalsTexAntiAliasingParam);

	cgGLDisableProfile(RenderState::sCgFragmentProfile);

	PrintGLerror("antialiasing");
}


void DeferredShader::FirstPassShadow(FrameBufferObject *fbo, 
									 const Matrix4x4 &matViewing, 
									 ShadowMapping *shadowMapping)
{
	GLuint colorsTex = fbo->GetColorBuffer(0)->GetTexture();
	GLuint positionsTex = fbo->GetColorBuffer(1)->GetTexture();
	GLuint normalsTex = fbo->GetColorBuffer(2)->GetTexture();
	GLuint shadowMap = shadowMapping->mFbo->GetDepthTex();

	static Matrix4x4 biasMatrix(0.5f, 0.0f, 0.0f, 0.0f,
								0.0f, 0.5f, 0.0f, 0.0f,
								0.0f, 0.0f, 0.5f, 0.0f,
								0.5f, 0.5f, 0.5f, 1.0f); //bias from [-1, 1] to [0, 1]

	Matrix4x4 inverseView = matViewing;
	inverseView.Invert();

	Matrix4x4 lightProjView = shadowMapping->mLightViewMatrix * shadowMapping->mLightProjectionMatrix;

	//cout << "matview:\n" << matViewing << endl;
	//cout << "proj:\n" <<  shadowMapping->mLightProjectionMatrix << endl;
	//cout << "view:\n" <<  shadowMapping->mLightViewMatrix << endl;

	// compute combined matrix that transforms pixels into light texture space
	/*Matrix4x4 shadowMatrix = biasMatrix * 
							 shadowMapping->mLightProjectionMatrix * 
							 shadowMapping->mLightViewMatrix *
							 inverseView;
*/
	Matrix4x4 shadowMatrix = //inverseView * 
							 lightProjView * 
							 biasMatrix;
/*
Matrix4x4 shadowMatrix = biasMatrix * 
							 lightProjView; 
							 //inverseView;
*/
							 cout << "combined:\n" <<  shadowMatrix << endl;

	fbo->Bind();
	glDrawBuffers(1, mymrt + 3);

	
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	cgGLEnableProfile(RenderState::sCgFragmentProfile);

	cgGLBindProgram(sCgDeferredShadowProgram);

	cgGLSetTextureParameter(sColorsTexShadowParam, colorsTex);
	cgGLEnableTextureParameter(sColorsTexShadowParam);

	cgGLSetTextureParameter(sPositionsTexShadowParam, positionsTex);
	cgGLEnableTextureParameter(sPositionsTexShadowParam);

	cgGLSetTextureParameter(sNormalsTexShadowParam, normalsTex);
	cgGLEnableTextureParameter(sNormalsTexShadowParam);
	
	cgGLSetTextureParameter(sShadowMapParam, shadowMap);
	cgGLEnableTextureParameter(sShadowMapParam);

	cgGLSetMatrixParameterfc(sShadowMatrixParam, (const float *)shadowMatrix.x);

	glColor3f(1.0f, 1.0f, 1.0f);
	
	glBegin(GL_QUADS);

	float offs2 = 0.5f;

	glTexCoord2f(0, 0); glVertex3f(-offs2, -offs2, -0.5f);
	glTexCoord2f(1, 0); glVertex3f( offs2, -offs2, -0.5f);
	glTexCoord2f(1, 1); glVertex3f( offs2,  offs2, -0.5f);
	glTexCoord2f(0, 1); glVertex3f(-offs2,  offs2, -0.5f);

	glEnd();

	cgGLDisableTextureParameter(sColorsTexShadowParam);
	cgGLDisableTextureParameter(sPositionsTexShadowParam);
	cgGLDisableTextureParameter(sNormalsTexShadowParam);
	cgGLDisableTextureParameter(sShadowMapParam);

	cgGLDisableProfile(RenderState::sCgFragmentProfile);
	
	FrameBufferObject::Release();

	PrintGLerror("deferred shading");
}



} // namespace
